--
-- HeapTipTrigger
--
-- M@D Author:  Heady
-- M@D date: 13.06.2011
-- M@D update LS13: 12.11.2013
-- M@D last update: 20.08.2014
--
-- > Copyright (C) Heady - www.planet-ls.de < --
--

print("  HeapTipTrigger v13.2");

local modDir = g_currentModDirectory;
g_currentModDirectoryOfHeapTipTrigger = modDir.."heapTipTrigger.xml";
BaseMission.modDirectoryOfHeapTipTrigger = modDir.."heapTipTrigger.xml";

if BaseMission.modHeapTipTriggerAddon == nil then
	BaseMission.modHeapTipTriggerAddon = {};
	print("  Register Variable: BaseMission.modHeapTipTriggerAddon");
end;


HeapTipTrigger = {};

function HeapTipTrigger.prerequisitesPresent(specializations)
    return true;
end;

--example: if g_currentMission.modDirectoryOfHeapTipTrigger ~= nil then g_currentMission:createNewHeapTipTrigger(x, 0.5, z, yr);

function BaseMission:createNewHeapTipTrigger(positionX, offsetY, positionZ, yRot)
	--print("positionX "..positionX.." offsetY "..offsetY.." positionZ "..positionZ.." yRot "..yRot)
	if positionX ~= nil or positionZ ~= nil then	
		local yOffset = Utils.getNoNil(offsetY, 0.5);
		local yr = Utils.getNoNil(yRot, math.random(0, 2*math.pi));
		--local yr = Utils.getNoNil(yRot, math.random(-(math.pi/2), (math.pi/2))*math.pi);
		--print("rad me ", math.random(-(math.pi/2), (math.pi/2))*math.pi, "rad ", math.random()*math.pi)

		if g_server ~= nil then
			local vehicle = g_currentMission:loadVehicle(g_currentModDirectoryOfHeapTipTrigger, positionX, yOffset, positionZ, yr);
			return vehicle; --only possible return on server; if you need it on client, you must make a send event on your script
		else
			g_client:getServerConnection():sendEvent(HeapTipTriggerCreateNewMPEvent:new(positionX, yOffset, positionZ, yr));
		end;
	else
		print("HeapTipTrigger not created, positionX or positionZ is nil");
	end;
	
end;

function HeapTipTrigger:load(xmlFile)

	self.setFillLevel = Utils.overwrittenFunction(self.setFillLevel, HeapTipTrigger.setFillLevel);
	self.tipTriggerCallback = HeapTipTrigger.tipTriggerCallback;
	self.updateMoving = HeapTipTrigger.updateMoving;
	self.shovelGetCallback = HeapTipTrigger.shovelGetCallback;
	self.getTipInfoForTrailer = HeapTipTrigger.getTipInfoForTrailer;
	self.getNoAllowedText = HeapTipTrigger.getNoAllowedText;
	self.updateTrailerTipping = HeapTipTrigger.updateTrailerTipping;
	

	g_currentMission:addTipTrigger(self);

	self.triggerId = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.tipTrigger#index"));
	addTrigger(self.triggerId, "tipTriggerCallback", self);


	-- to be detected by raycast callbacks, e.g. by the forrest mod wood crusher
	g_currentMission:addNodeObject(self.triggerId, self);
	
	if self.priceMultipliers == nil or self.acceptedFillTypes == nil then
        self.acceptedFillTypes = {};
        self.priceMultipliers = {};

        local fillTypes = getXMLString(xmlFile, "vehicle.fillTypes#fruitTypes");
        if fillTypes ~= nil then
            local types = Utils.splitString(" ", fillTypes);
            for k,v in pairs(types) do
                local desc = Fillable.fillTypeNameToDesc[v];
                if desc ~= nil then
                    self.acceptedFillTypes[desc.index] = true;
                    self.priceMultipliers[desc.index] = 0;
                end;
            end;
        end;
    end
	
	self.moveNode = Utils.indexToObject(self.components, getXMLString(xmlFile, "vehicle.fillPlanes#moveNode"));
	addTrigger(self.moveNode, "shovelGetCallback", self);
	
	self.attachableID = nil;
	self.trailersInTrigger = {};
	
	self.triggerStartId = self.triggerId;
	self.triggerTipWidth = math.huge;

	self.allowToRemove = false;
	self.removeTime = 2000;
	
	self.waitToStaticTime = 0;
	self.staticPos = {0, 0, 0};
	self.setStatic = true;
	

	-- add from heaptiptrigger addon	
	for i=1, table.getn(g_currentMission.modHeapTipTriggerAddon) do
		local addonModDir = g_currentMission.modHeapTipTriggerAddon[i];
		local addonxmlFile = loadXMLFile("TempConfig", addonModDir.."HeapTipTriggerAddon.xml");
		
		local fillTypes = getXMLString(addonxmlFile, "addon.addFillTypes#fillTypes");
        if fillTypes ~= nil then
            local types = Utils.splitString(" ", fillTypes);
            for k,v in pairs(types) do
                local desc = Fillable.fillTypeNameToDesc[v];
                if desc ~= nil then
                    self.acceptedFillTypes[desc.index] = true;
                    self.priceMultipliers[desc.index] = 0;
                end;
            end;
        end;
		
        local i = 0;
        while true do
            local key = string.format("addon.fillPlanes.fillPlane(%d)", i);
            if not hasXMLProperty(addonxmlFile, key) then
                break;
            end;
            local fillPlane = {};
            fillPlane.nodes = {};
            local fillType = getXMLString(addonxmlFile, key.."#type");
            if fillType ~= nil then
                local nodeI = 0;
                while true do
                    local nodeKey = key..string.format(".node(%d)", nodeI);
                    if not hasXMLProperty(addonxmlFile, nodeKey) then
                        break;
                    end;
					
					local i3dNode = Utils.loadSharedI3DFile("HeapTipTriggerAddon.i3d", addonModDir);
					local node = getChildAt(i3dNode, getXMLInt(addonxmlFile, nodeKey.."#index"))
					link(self.moveNode, node);
					delete(i3dNode);
							
                    if node ~= nil then
                        local defaultX, defaultY, defaultZ = getTranslation(node);
                        local defaultRX, defaultRY, defaultRZ = getRotation(node);
                        setVisibility(node, false);

                        local animCurve = AnimCurve:new(linearInterpolatorTransRotScale);
                        local keyI = 0;
                        while true do
                            local animKey = nodeKey..string.format(".key(%d)", keyI);
                            local keyTime = getXMLFloat(addonxmlFile, animKey.."#time");
                            local x,y,z = Utils.getVectorFromString(getXMLString(addonxmlFile, animKey.."#translation"));
                            if y == nil then
                                y = getXMLFloat(addonxmlFile, animKey.."#y");
                            end;
                            local rx,ry,rz = Utils.getVectorFromString(getXMLString(addonxmlFile, animKey.."#rotation"));
                            local sx,sy,sz = Utils.getVectorFromString(getXMLString(addonxmlFile, animKey.."#scale"));
                            if keyTime == nil then
                                break;
                            end;
                            local x = Utils.getNoNil(x, defaultX);
                            local y = Utils.getNoNil(y, defaultY);
                            local z = Utils.getNoNil(z, defaultZ);
                            local rx = Utils.getNoNil(rx, defaultRX);
                            local ry = Utils.getNoNil(ry, defaultRY);
                            local rz = Utils.getNoNil(rz, defaultRZ);
                            local sx = Utils.getNoNil(sx, 1);
                            local sy = Utils.getNoNil(sy, 1);
                            local sz = Utils.getNoNil(sz, 1);
                            animCurve:addKeyframe({x=x, y=y, z=z, rx=rx, ry=ry, rz=rz, sx=sx, sy=sy, sz=sz, time = keyTime});
                            keyI = keyI +1;
                        end;
                        if keyI == 0 then
                            local minY, maxY = Utils.getVectorFromString(getXMLString(addonxmlFile, nodeKey.."#minMaxY"));
                            local minY = Utils.getNoNil(minY, defaultY);
                            local maxY = Utils.getNoNil(maxY, defaultY);
                            animCurve:addKeyframe({x=defaultX, y=minY, z=defaultZ, rx=defaultRX, ry=defaultRY, rz=defaultRZ, sx=1, sy=1, sz=1, time = 0});
                            animCurve:addKeyframe({x=defaultX, y=maxY, z=defaultZ, rx=defaultRX, ry=defaultRY, rz=defaultRZ, sx=1, sy=1, sz=1, time = 1});
                        end;
                        table.insert(fillPlane.nodes, {node=node, animCurve = animCurve});
                    end;
                    nodeI = nodeI +1;
                end;
                if table.getn(fillPlane.nodes) > 0 then
                    self.fillPlanes[fillType] = fillPlane;
                end;
            end;
            i = i +1;
        end;
		
		delete(addonxmlFile);
	
	end;

end;

function HeapTipTrigger:delete()

	g_currentMission:removeTipTrigger(self);
    removeTrigger(self.triggerId);
	
	removeTrigger(self.moveNode);
	
	if g_currentMission ~= nil then	
		for k,v in pairs(g_currentMission.trailerTipTriggers) do
			for key, tipTrigger in pairs(v) do
				if tipTrigger == self then
					v[key] = nil;
				end;
			end;
		end;
	end;
	
end;

function HeapTipTrigger:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)

	if self.fillLevel <= 0 then
		self.allowToRemove = true;
	else
		setRigidBodyType(self.rootNode, "Static"); 
		self.setStatic = false;
		self:setFillLevel(self.fillLevel, self.currentFillType);
	end;
	
	return BaseMission.VEHICLE_LOAD_OK;
end;

function HeapTipTrigger:mouseEvent(posX, posY, isDown, isUp, button)
end;

function HeapTipTrigger:keyEvent(unicode, sym, modifier, isDown)
end;

function HeapTipTrigger:readStream(streamId, connection)
    if connection:getIsServer() then
        if self.synchronizeFillLevel then
            local fillLevel = streamReadFloat32(streamId);
            local fillType = streamReadUIntN(streamId, Fillable.sendNumBits);
			self.setStatic = false;
            self:setFillLevel(fillLevel, fillType);
        end
    end
end;

function HeapTipTrigger:writeStream(streamId, connection)
    if not connection:getIsServer() then
        if self.synchronizeFillLevel then
            streamWriteFloat32(streamId, self.fillLevel);
            streamWriteUIntN(streamId, self.currentFillType, Fillable.sendNumBits);
        end
    end;
end;

function HeapTipTrigger:update(dt)

	self.waitToStaticTime = self.waitToStaticTime + dt; -- realtime ms	
	
	local x, y, z = getTranslation(self.components[1].node);
	
	if self.setStatic and self.waitToStaticTime > 100 then
		local x1, y1, z1 = unpack(self.staticPos);
		if x1 == x and y1 == y and z1 == z then
			setRigidBodyType(self.rootNode, "Static");
			self.setStatic = false;
		end;
		self.staticPos = {x, y, z};
		self.waitToStaticTime = 0;
	end;

end;

function HeapTipTrigger:updateTick(dt)

	if self.isServer then
	
		if self.attachableID ~= nil and self.fillLevel > 0 then		
			if self.attachableID.isShovel ~= nil or self.attachableID.getFromSilo ~= nil or self.attachableID.shovelTipReferenceNode ~= nil then -- siloschaufel or mischwagen etc or standard schaufel
				if self.attachableID.fillLevel < self.attachableID.capacity then
					local m = self.attachableID.capacity * (dt/1000);
					local deltaFillLevel = math.min(m, self.fillLevel);
					self.attachableID:setFillLevel(self.attachableID.fillLevel+deltaFillLevel, self.currentFillType);
					self:setFillLevel(self.fillLevel-deltaFillLevel, self.currentFillType);
					self.allowToRemove = true;	
				end;
			end;
		end;
		
		if self.allowToRemove then			
			if self.fillLevel <= 0 then
				self.removeTime = self.removeTime - dt;
				if self.removeTime <= 0 then
					g_currentMission.vehiclesToDelete[self] = self;
				end;
			end;
		end;
		
	end;

end;

function HeapTipTrigger:draw()
end;

function HeapTipTrigger:updateTrailerTipping(trailer, fillDelta, fillType)


	--local fillType = self.currentFillType;
	--if self.fillLevel <= 0 then
		--fillType = self.trailersInTrigger[1].currentFillType;
	--end;
	
	local oldFillLevel = self.fillLevel - fillDelta;
	self.fillLevel = math.min(oldFillLevel, self.capacity);	
	
	self:setFillLevel(self.fillLevel, fillType);
	
	local refill = oldFillLevel - self.fillLevel;
	if refill > 0 then
		for i=1, table.getn(self.trailersInTrigger) do
			local trailer = self.trailersInTrigger[i];
			--trailer:setFillLevel(trailer.fillLevel + refill, trailer.currentFillType);
			trailer:onEndTip();
		end;
	end;

	--self.allowToRemove = false;	

end;


function HeapTipTrigger:tipTriggerCallback(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)

	if onEnter then
		local trailer = g_currentMission.objectToTrailer[otherShapeId];
		if trailer ~= nil and trailer.allowTipDischarge then
			if g_currentMission.trailerTipTriggers[trailer] == nil then
				g_currentMission.trailerTipTriggers[trailer] = {};
			end;
			table.insert(g_currentMission.trailerTipTriggers[trailer], self);
			table.insert(self.trailersInTrigger, trailer);
		end;
	elseif onLeave then
		local trailer = g_currentMission.objectToTrailer[otherShapeId];
		if trailer ~= nil and trailer.allowTipDischarge then
			local triggers = g_currentMission.trailerTipTriggers[trailer];
			if triggers ~= nil then
				for i=1, table.getn(triggers) do
					if triggers[i] == self then
						table.remove(triggers, i);
						if table.getn(triggers) == 0 then
							g_currentMission.trailerTipTriggers[trailer] = nil;
						end;
						break;
					end;
				end;
			end;
			for i=1, table.getn(self.trailersInTrigger) do
				if self.trailersInTrigger[i] == trailer then
					table.remove(self.trailersInTrigger, i);
				end;
			end;
		end;
	end;

end;

function HeapTipTrigger:shovelGetCallback(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)

    if onEnter then
		if otherId ~= nil then	
			local attachable = g_currentMission.nodeToVehicle[otherId];
			if attachable ~= nil then
				if attachable.fillTypes ~= nil and attachable.setFillLevel ~= nil and attachable.fillLevel ~= nil then
					if attachable.fillLevel < attachable.capacity and attachable:allowFillType(self.currentFillType, true) then
						self.attachableID = attachable;	
					end;
				end;
			end;
		end;		
	elseif onLeave then
		if otherId ~= nil then	
			local attachable = g_currentMission.nodeToVehicle[otherId];
			if attachable ~= nil then
				if attachable == self.attachableID then
					self.attachableID = nil;
				end;
			end;
		end;					
	end;	
	
end;

function HeapTipTrigger:setFillLevel(superFunc, fillLevel, fillType, force)
    if (force == nil or force == false) and not self:allowFillType(fillType, false) then
        return
    end;

    self.currentFillType = fillType;
    self.fillLevel = fillLevel;
    if self.fillLevel > self.capacity then
        self.fillLevel = self.capacity;
    end;
    if self.fillLevel < 0 then
        self.fillLevel = 0;
        self.currentFillType = Fillable.FILLTYPE_UNKNOWN;
    end;


    if self.isClient then
        if self.currentFillPlane ~= nil then
            for _, node in ipairs(self.currentFillPlane.nodes) do
                setVisibility(node.node, false);
            end;
            self.currentFillPlane = nil;
        end;
        if self.fillPlanes ~= nil and self.defaultFillPlane ~= nil and fillType ~= Fillable.FILLTYPE_UNKNOWN then
            local fillTypeName = Fillable.fillTypeIntToName[fillType];
            local fillPlane = self.fillPlanes[fillTypeName];
            if fillPlane == nil then
                fillPlane = self.defaultFillPlane;
            end;
            local t = self.fillLevel/self.capacity;
            for _, node in ipairs(fillPlane.nodes) do
                local x,y,z, rx,ry,rz, sx,sy,sz = node.animCurve:get(t);
				
                setTranslation(self.moveNode, x, y, z);
                setRotation(self.moveNode, rx, ry, rz);
                setScale(node.node, sx, sy, sz);
                setVisibility(node.node, self.fillLevel > 0 and not self.setStatic);
            end;
            self.currentFillPlane = fillPlane;
        end;
    end;
	
	
end;

function HeapTipTrigger:getTipInfoForTrailer(trailer, tipReferencePointIndex)
	local minDistance, bestPoint = TipTrigger.getTipDistanceFromTrailer(self, trailer, tipReferencePointIndex);
	
	--local minDistance = 0;
	--local bestPoint = trailer.tipReferencePoints[1];

	local isAllowed = false;
	if self.currentFillType == trailer.currentFillType or self.fillLevel == 0 then
		isAllowed = true;
	end;
 
	return isAllowed, minDistance, bestPoint;
end;

function HeapTipTrigger:getNoAllowedText(trailer)
	if trailer.currentFillType ~= Fillable.FILLTYPE_UNKNOWN and self.fillLevel > 0 and self.currentFillType ~= trailer.currentFillType then
		return Fillable.fillTypeIndexToDesc[trailer.currentFillType].nameI18N .. g_i18n:getText("notAcceptedHere");
	end;
	
	return nil;
end;



if BaseMission.modHeapTipTrigger == nil then
	BaseMission.modHeapTipTrigger = true;
	print("  Register Variable: BaseMission.modHeapTipTrigger");
	
	
	-- TRAILER --
	local TrailerUpdate = Trailer.update;
	Trailer.update = function(self, dt)

		if self.printStrg == nil then
			self.printStrg = false;
		end;

		if self:getIsActiveForInput() then
			if self.tipReferencePoints ~= nil then
				if self.tipReferencePoints[1] ~= nil then
		
					self.printStrg = false;
				
					if g_currentMission.trailerInTipRange == nil and ( self.tipState == Trailer.TIPSTATE_CLOSED or self.tipState == Trailer.TIPSTATE_CLOSING ) then --and not self.create_HeapTipTrigger then
						if Input.isKeyPressed(Input.KEY_lctrl) and InputBinding.hasEvent(InputBinding.ACTIVATE_OBJECT) then
							local x,y,z = getWorldTranslation(self.tipReferencePoints[1].node);
							local xr,yr,zr = getWorldRotation(self.tipReferencePoints[1].node);
							g_currentMission:createNewHeapTipTrigger(x, 0.5, z, yr);
						end;
						if Input.isKeyPressed(Input.KEY_lctrl) then
							self.printStrg = true;
						else
							self.printStrg = false;
						end;
					end;
					
				end;
			end;			
		end;

		TrailerUpdate(self, dt);
	end;

	local TrailerDraw = Trailer.draw;
	Trailer.draw = function(self)

		if self.printStrg then
			g_currentMission:addExtraPrintText("HeapTipTrigger erstellen: STRG + "..InputBinding.getKeyNamesOfDigitalAction(InputBinding.ACTIVATE_OBJECT));
		end;	

		TrailerDraw(self);
	end;
	
	
	-- COMBINE --	
	local CombineUpdate = Combine.update;
	Combine.update = function(self, dt)

		if self.printStrg == nil then
			self.printStrg = false;
		end;

		if self:getIsActiveForInput() then
			if self.pipeRaycastNode ~= nil then		
				self.printStrg = false;
			
				if self.pipeStateIsUnloading[self.currentPipeState] and self:getIsPipeUnloadingAllowed() then
					if Input.isKeyPressed(Input.KEY_lctrl) and InputBinding.hasEvent(InputBinding.ACTIVATE_OBJECT) then
						local x,y,z = getWorldTranslation(self.pipeRaycastNode);
						local xr,yr,zr = getWorldRotation(self.pipeRaycastNode);
						g_currentMission:createNewHeapTipTrigger(x, 0.5, z, yr);
					end;
					if Input.isKeyPressed(Input.KEY_lctrl) then
						self.printStrg = true;
					else
						self.printStrg = false;
					end;
				end;
			end;			
		end;

		CombineUpdate(self, dt);
	end;

	local CombineDraw = Combine.draw;
	Combine.draw = function(self)

		if self.printStrg then
			g_currentMission:addExtraPrintText("HeapTipTrigger erstellen: STRG + "..InputBinding.getKeyNamesOfDigitalAction(InputBinding.ACTIVATE_OBJECT));
		end;	

		CombineDraw(self);
	end;
	
end;



HeapTipTriggerCreateNewMPEvent = {};
HeapTipTriggerCreateNewMPEvent_mt = Class(HeapTipTriggerCreateNewMPEvent, Event);

InitEventClass(HeapTipTriggerCreateNewMPEvent, "HeapTipTriggerCreateNewMPEvent");

function HeapTipTriggerCreateNewMPEvent:emptyNew()
    local self = Event:new(HeapTipTriggerCreateNewMPEvent_mt);
    self.className="HeapTipTriggerCreateNewMPEvent";
    return self;
end;

function HeapTipTriggerCreateNewMPEvent:new(positionX, yOffset, positionZ, yr)
    local self = HeapTipTriggerCreateNewMPEvent:emptyNew()
    self.positionX = positionX;
    self.yOffset = yOffset;
    self.positionZ = positionZ;
    self.yr = yr;
    return self;
end;

function HeapTipTriggerCreateNewMPEvent:readStream(streamId, connection)
	--local id = streamReadInt32(streamId)
    --self = networkGetObject(id);
	
	self.positionX = streamReadInt32(streamId);
	self.yOffset = streamReadInt32(streamId);
	self.positionZ = streamReadInt32(streamId);
	self.yr = streamReadInt32(streamId);

	--self.object.create_HeapTipTrigger = streamReadBool(streamId);
	g_currentMission:createNewHeapTipTrigger(self.positionX, self.yOffset, self.positionZ, self.yr);
end;

function HeapTipTriggerCreateNewMPEvent:writeStream(streamId, connection)
	--streamWriteInt32(streamId, networkGetObjectId(self));
	
	streamWriteInt32(streamId, self.positionX);
	streamWriteInt32(streamId, self.yOffset);
	streamWriteInt32(streamId, self.positionZ);
	streamWriteInt32(streamId, self.yr);
	--streamWriteBool(streamId, self.object.create_HeapTipTrigger);
end;
